(* CS 655 - Spring 2006 (NF Skeleton Version 3) 
 * 
 * YOUR NAME HERE (write your name here so that if the files get mixed up
 * somehow I can still give you credit for your work)
 *  
 * Note that this is not the world's greatest example of software
 * engineering, encapsulation, or data hiding. To pick just one example, in
 * a real implementation you would want to abstract the communication with
 * the theorem prover so that other theorem provers (e.g., CVC) could be
 * plugged in. This is just a classroom exercise. That said, it run on
 * real C programs. 
 *)

(* This is the main file in our null-finding analysis. *) 

open Cil
open Printf

(*************************************************************************** 
 * SECTION 0 -- INPUT ASSUMPTIONS
 *
 * These comments describe ways in which CIL and its various extensions
 * have boiled down all of the possible forms you would expect to see in a
 * C AST into just a few. It will look quite a bit like IMP + Gotos when you
 * finally get around to it. (surprise!). 
 ***************************************************************************)
(* The AST itself is in cil/src/cil.mli 
 *
 * However, we will transform and simplify the full AST before its gets to
 * you, so you can make the following simplifying assumptions:
 * (1) You will not see switch/break/default/continue statements -- they
 * will be replaced by ifs and gotos. 
 * (2) Each lvalue will contain at most one Mem() constructor. Instead of
 * "**x =  y" you'll see "temp_55 = *x ; *temp_55 = y;" 
 * In addition, you'll get all the benefits of the Simplify module (not to
 * be confused with the Simplify theorem prover ...): 
  
1. All expressions are either 

 basic::=
    Const _ 
    Addrof(Var v, NoOffset)
    StartOf(Var v, NoOffset)
    Lval(Var v, off), where v is a variable whose address is not taken
                      and off contains only "basic"

 exp::=
    basic
    Lval(Mem basic, NoOffset)
    BinOp(bop, basic, basic)
    UnOp(uop, basic)
    CastE(t, basic)
   
 lval ::= 
    Mem basic, NoOffset
    Var v, off, where v is a variable whose address is not taken and off
                contains only "basic"

 - all sizeof and alignof are turned into constants
 - accesses to variables whose address is taken is turned into "Mem" accesses
 - same for accesses to arrays
 - all field and index computations are turned into address arithmetic, 
   including bitfields.
*) 
(* 
 * Want to do a context-sensitive analysis? Check out
 * cil/src/ext/callgraph.mli for something to compute the static call
 * graph.
 *)

(*************************************************************************** 
 * These are some global configuration settings that are initialized in 
 * main() below. 
 ***************************************************************************)

let be_quiet = ref false (* when true, we only print out error messages *) 
let debug_tp = ref false (* when true, we print out theorem prover calls *)
let error_count = ref 0  (* # errors reported so far *)
let stop_at_error = ref (20) (* stop when we get to this many errors *) 
let be_certain = ref false (* only report errors that are likely real *) 
let only_functions = ref [] (* only analyze these functions *) 

(*************************************************************************** 
 * These two functions are used for debugging and error reporting. The
 * file cil/ocamlutil/pretty.mli is also used for pretty-printing C 
 * or CIL structures. 
 ***************************************************************************)
(* 
 * This "debug" function works like printf, but only emits output if
 * !be_quiet is not set. 
 *) 
let debug fmt =  
  let k result = 
    if not !be_quiet then begin
      output_string stdout result ;
      flush stdout 
    end 
  in 
  Printf.kprintf k fmt 

let report_error (path : location list) fmt = 
  let k result = 
    output_string stdout ("\n" ^ result) ;
    List.iter (fun loc ->
      printf "-> %s:%d " loc.file loc.line 
    ) (List.rev path) ; 
    output_string stdout ("\n" ^ result  ^"\n");
    flush stdout ;
    incr error_count ;
    if !error_count = !stop_at_error then begin
      output_string stdout "Maximal Error Count Reached! Exiting!\n\n" ;
      exit 1 
    end 
  in 
  Printf.kprintf k fmt 

(*************************************************************************** 
 * SECTION A -- SYMBOLIC STATE 
 *
 * We will use Cil's "C Expressions" as our symbolic expressions for
 * symbolic execution. We'll use special variables to represent things 
 * we have only symbolic knowledge of. We will use the McCarrthy
 * Select/Update "Memory Is A List Of Writes" formalism for modeling the
 * heap. 
 ***************************************************************************)
type symexp = Cil.exp 

type symmem = (symexp * symexp) list (* = (address, value_written) list *) 

module OrderedString =
  struct
    type t = string
    let compare = compare
  end
module OrderedInt =
  struct
    type t = int
    let compare = compare
  end
module StringMap = Map.Make(OrderedString)
module IntSet = Set.Make(OrderedInt)

(*************************************************************************** 
 * Our Symbolic State is purely applicative -- no side effects. It includes
 * all of the usual suspects:
 *  sigma     maps variables to values
 *  mu        is a bit list of memory writes that you can use to figure
 *            out what *p should evaluate to
 *  visited   keeps track of where we have been so that we can make sure
 *            that we don't loop forever
 *  assumptions -- facts that we have learned are true along this path
 *  path      this particular execution path (in reverse order) -- used
 *            to print out error traces 
 ***************************************************************************)
(* state is purely applicative (functional) -- no side effects *) 
type state = {
  sigma : symexp StringMap.t ; (* maps strings (variable names) to symexps *)
  mu : symmem ;
  visited : IntSet.t ; (* which statements have we visited? *) 
  assumptions : symexp list ; 
    (* things we have assumed to be true along this path *) 
  path : location list ; (* how did we get here? *) 
} 

(* a null state containing no information *) 
let empty_state = {
  sigma = StringMap.empty ;
  mu = [] ;
  visited = IntSet.empty ;
  assumptions = [] ; 
  path = [] ; 
} 

(* prints out a state to the screen so that you can figure out what is
 * going on *) 
let debug_state state =
  StringMap.iter (fun k v ->
    let estr = Pretty.sprint 80 (defaultCilPrinter#pExp () v) in
    debug "Sigma %20s = %s\n" k estr
  ) state.sigma ;
  List.iter (fun (a,v) ->
    let astr = Pretty.sprint 80 (defaultCilPrinter#pExp () a) in
    let vstr = Pretty.sprint 80 (defaultCilPrinter#pExp () v) in
    debug "Mu    %20s -> %s\n" astr vstr
  ) state.mu ; 
  List.iter (fun a ->
    let astr = Pretty.sprint 80 (defaultCilPrinter#pExp () a) in
    debug "Assume %s\n" astr
  ) state.assumptions ; 
  () 

(* returns a new state in which 'value' has been written to memory heap
 * location 'address' *)
let update_memory state address value =
  { state with mu = (address,value) :: state.mu } 

(* returns a new state in which variable 'varinfo' has been assigned
 * the new value 'exp' *)
let assign old_state varinfo exp = 
  let old_sigma = old_state.sigma in 
  { old_state with sigma = StringMap.add varinfo.vname exp old_sigma }

(* looks up the value of 'varinfo' in the state. May raise Not_found *)
let lookup state varinfo = 
  StringMap.find varinfo.vname state.sigma 

(* returns true if we have already been to this statement *) 
let already_visited state stmt =
  IntSet.mem stmt.sid state.visited 

(* returns a new state in which we have visited 'stmt' *) 
let mark_as_visited old_state stmt = 
  { old_state with visited = IntSet.add stmt.sid old_state.visited } 

(* returns a new state in which 'newly_assumed' is known to be true *) 
let assume old_state newly_assumed = 
  { old_state with assumptions = newly_assumed :: old_state.assumptions }  

(* returns a new state in which we remember that we have come from
 * 'location' so that we can print out an error trace later *) 
let add_to_path state location = 
  match state.path with
  | last_visit :: rest_of_path when last_visit = location -> state
  | _ -> { state with path = location :: state.path } 

(* We will often need to make a fresh symbolic value that we know nothing
 * about (this is like the \forall x. ... in the notes) ... "fresh_value"
 * does that for us *) 
let value_counter = ref 0 
let fresh_value ?va () = 
  let str = 
    match va with
    | None -> "|" 
    | Some(va) -> "|" ^ va.vname 
  in
  let c = !value_counter in
  incr value_counter ;
  let va = makeVarinfo false (Printf.sprintf "%s%d" str c) (TVoid([])) in
  Lval(Var(va),NoOffset)

(* this function throws away all information we have about the values 
 * of local variables and heap values -- this is typically a conservative
 * "option of last resort" when something happens that we don't know how to
 * model *) 
let throw_away_state old_state =
  let new_sigma = StringMap.mapi (fun old_key old_val -> 
    fresh_value () 
  ) old_state.sigma in
  { old_state with sigma = new_sigma ; mu = [] ; assumptions = [] } 

(* returns true if the given expression represents one of our fresh,
 * unknown symbolic values *) 
let is_unknown_symexp e = match e with
  | Lval(Var(va),NoOffset) when va.vname.[0] = '|' -> true 
  | _ -> false 

(* this convenience function returns the symbolic expression (and 
 * C/CIL expression) associated with 'true' or 'false'. Recall that in 
 * C we have "false == 0" and "true <> 0". *)
let se_of_bool b = match b with
  | true -> Const(CInt64(Int64.one,IInt,None))
  | false -> Const(CInt64(Int64.zero,IInt,None))

(*************************************************************************** 
 * SECTION B -- THEOREM PROVING 
 *
 * We will make calls to Simplify theorem prover to decide if a certain
 * symbolic expression must be true or not. For example, given that we know
 * "x0 = 55", we should be able to reason that "y = x0 ; y > 100" is false. 
 * The theorem prover is part of that. 
 *
 * The theorem prover interface also encapsulates some basic knowledge
 * about C programs. For example:
 *
 * (1) String constants like "hello" are never zero.
 * (2) Character constants besides '\000' are not zero.
 * (3) In C, "if (x)" really means "if (x != 0)".
 * (4) &variable is never zero. 
 * (5) &v1 != &v2 
 *
 * This sounds cool, but in reality the Theorem prover will not be able to
 * solve everything for you. It will often return "Unknown" -- for example,
 * in questions involving multiplication or division. Also, it doesn't
 * understand bitwise operations. 
 ***************************************************************************)

(* theorem provers are incomplete, so often "Unknown" is the answer *) 
type tp_result = 
  | Must_Be_True
  | Must_Be_False
  | Unknown

(* Uses an external theorem prover to try to decide if "e" is necessarily
 * true. *) 
let valid_regexp = Str.regexp ".*Valid\\." 
let invalid_regexp = Str.regexp ".*Invalid\\." 
let bad_regexp = Str.regexp ".*Bad input" 

(* This powerful procedure asks the Theorem Prover to decide if "e" 
 * must be true. If anything goes wrong, we return Unknown. *) 
let decide (s:state) e : tp_result = 
  match e with
  (* for constants, we don't even need to ask the theorem prover --
   * we can decide it locally *) 
  | Const(CInt64(x,_,_)) -> 
    if x = Int64.zero then Must_Be_False else Must_Be_True
  | Const(CStr(_)) 
  | Const(CWStr(_)) -> Must_Be_True
  | Const(CChr(c)) -> if c = '\000' then Must_Be_False else Must_Be_True
  | _ -> begin 

  (* if the expression is already a relational operator, we don't need
   * to coerce on the implicit "!= 0" suffix *) 
  let is_relop bop = match bop with
  | Lt                                  (** <  (arithmetic comparison) *)
  | Gt                                  (** >  (arithmetic comparison) *)  
  | Le                                  (** <= (arithmetic comparison) *)
  | Ge                                  (** >  (arithmetic comparison) *)
  | Eq                                  (** == (arithmetic comparison) *)
  | Ne                                  (** != (arithmetic comparison) *)
  | LAnd                                (** logical and. *) 
  | LOr                                 (** logical or. *) 
  -> true
  | _ -> false 
  in 
  let must_coerce e = match e with
  | BinOp(bop, _, _, _) when is_relop bop -> false
  | UnOp(LNot,_,_) -> false 
  | _ -> true 
  in

  (* Now we want to convert the CIL AST expression into a Simplify Query
   * String. Simplify expects queries like
   *
   * (IMPLIES (= X 5) (>= (+ X 2) 7))
   *
   * However, you cannot just use program variables directly -- you might
   * have a C program variable named "IMPLIES", but that would confuse
   * Simplify. So we rename all of the variables involved to v0, v1, ...
   *)
  let vht = Hashtbl.create 255 in
  let vht_counter = ref 0 in 
  let handle_va va =
    try 
      Hashtbl.find vht va.vname
    with Not_found -> begin
      let res = Printf.sprintf "v%d" !vht_counter in
      incr vht_counter ;
      Hashtbl.add vht va.vname res ;
      res 
    end 
  in 
  (* We also introduce a special symbolic expression representing
   * the address of a variable. We know that addresses are all distinct and
   * they are all non-null. *) 
  let addrs_mentioned = Hashtbl.create 5 in 
  let handle_addrof va = 
    let res = "a" ^ (handle_va va) in 
    Hashtbl.replace addrs_mentioned res () ; 
    res
  in 

  (* This bit does a recursive walk to structurally convert a CIL
   * expression into a Simplify query. Note that C has implicit conversions
   * between bools and ints but that Simplify does not. Thus we insert 
   * " != 0" for int-like C expressions. *) 
  let rec to_str e want_bool = 
    let result = match e with
    | Const(CInt64(i64,_,_)) -> Int64.to_string i64
    | Const(CStr(_)) 
    | Const(CWStr(_)) 
    | Const(CChr(_)) 
    | Const(CReal(_,_,_)) -> failwith "don't know how ask simplify about this" 
    | AddrOf(Var(va),NoOffset) -> handle_addrof va 
    | AddrOf(lv) -> failwith "fixme -- I should handle &lval" 
    | Lval(Var(va),NoOffset) -> handle_va va
    | Lval(_) -> failwith "fixme -- I should handle other lvalues" 
    | UnOp(Neg,e,_) -> sprintf "(- 0 %s)" (to_str e false) 
    | UnOp(LNot,e,_) -> sprintf "(NOT %s)" (to_str e true) 
    | BinOp(bop,e1,e2,_) -> begin
      let bop_str, want_bool = match bop with
      | PlusA | PlusPI -> "+", false 
      | MinusA | MinusPI | MinusPP -> "-", false 
      | Mult -> "*", false 
      | Div -> "/", false 
      | Mod -> "%", false 
      | Lt -> "<", false 
      | Gt -> ">", false 
      | Le -> "<=" , false 
      | Ge -> ">=", false 
      | Ne -> "NEQ", false
      | Eq -> "EQ", false

      | LAnd -> "AND", true
      | LOr -> "OR", true
      | _ -> failwith "don't know how to ask simplify about this" 
      in
      sprintf "(%s %s %s)" bop_str (to_str e1 want_bool) 
        (to_str e2 want_bool) 
    end 
    | _ -> failwith "don't know how to ask simplify about this" 
    in 
    if want_bool && must_coerce e then 
      sprintf "(NEQ %s 0)" result 
    else
      result 
  in

  if not !be_quiet && !debug_tp then begin
    ignore (Pretty.printf "Theorem Prover Query: @[%a@]@!" 
      defaultCilPrinter#pExp e) ;
  end ; 

  try
    let str = to_str e true in
    (* we get to use all of the assumptions we have built up along this
     * path in order to help prove the query *) 
    let assumption_string_list = List.map (fun assumption -> 
      try (to_str assumption false) with _ -> "TRUE" 
    ) s.assumptions in
    (* we also get to assume that addresses are distinct and non-null *) 
    let distinct_assumption = 
      (Hashtbl.fold (fun name () str -> 
        str ^ name ^ " "
      ) addrs_mentioned "(DISTINCT 0 ") ^ ")"
    in 
    let make_query str = match 
      distinct_assumption :: assumption_string_list with
    | [] -> str
    | big_list -> 
      let b = Buffer.create 255 in
      bprintf b "(IMPLIES (AND" ;
      List.iter (fun elt -> bprintf b " %s" elt) big_list ;
      bprintf b ") %s)" str ;
      Buffer.contents b 
    in 
    let query str = 
      (if !debug_tp then debug "Theorem Prover Asking: %s\n" str) ; 
      try 
        (* in real life you do *NOT* want to spawn a new process every time
         * you make a theorem prover query -- however, for this homework
         * it inspires you not to make too many theorem prover calls *)
        let inchan, outchan = Unix.open_process "Simplify" in 
        output_string outchan (str ^ "\n") ; flush outchan ; 
        let finished = ref None in 
        let rec scan () = match !finished with
        | None -> 
          let reply = input_line inchan in 
          (if !debug_tp then debug "Theorem Prover Reply: %s\n" reply) ; 
          (if Str.string_match valid_regexp reply 0 then
            finished := Some(Must_Be_True)
          else if Str.string_match invalid_regexp reply 0 then
            finished := Some(Must_Be_False)
          else if Str.string_match bad_regexp reply 0 then begin 
            finished := Some(Unknown)
          end 
          ) ;
          scan () 
        | Some(x) -> 
          let _ = Unix.close_process (inchan,outchan) in 
          x
        in 
        scan () 
      with _ -> Unknown
    in 
    (* because Simplify is incomplete, it may say that both PREDICATE
     * and (NOT PREDICATE) can or cannot be proved. We only trust it when
     * it really knows. 
     *
     * As a concrete example (try it!) Simplify will say that both:
     * (IMPLIES (EQ X 5) (EQ ( * X X) 25))
     * and 
     * (IMPLIES (EQ X 5) (NOT (EQ ( * X X) 25)))
     * are Invalid (= not provable by Simplify).
     *)
    let q1 = make_query str in 
    let q2 = make_query (sprintf "(NOT %s)" str) in
    (match query q1, query q2 with
    | Must_Be_True, Must_Be_False -> Must_Be_True
    | Must_Be_False, Must_Be_True -> Must_Be_False
    | _, _ -> Unknown)
  with _ -> Unknown
  end 
  
let decide (s:state) e : tp_result = 
  let result = decide s e in 
  if not !be_quiet && !debug_tp then begin
    ignore (Pretty.printf "Theorem Prover Decide: @[%a@] = %s@!" 
      defaultCilPrinter#pExp e
      (match result with
      | Must_Be_True -> "true"
      | Must_Be_False -> "false"
      | _ -> "???")
      ) ;
  end ; 
  result 

(*************************************************************************** 
 * SECTION C -- Symbolic Expression Evaluation
 *
 * In CIL, expressions are "pure" (no side effects) and assignments happen
 * in Instructions. 
 *
 * We must be able to evaluate a CIL expression down into a symbolic
 * expression. For example, if "x = 5 && y = y_0" we want to turn
 * "x + y" into "5 + y_0". 
 *
 * This will involve handling more complicated expression forms, like *p. 
 ***************************************************************************)
(* 
 * function for turning a Cil expression "ce" into a symbolic expression
 * based on the state "s" 
 *)
let rec eval s ce = match ce with 
  | Lval(Var(va),NoOffset) -> 
    if is_unknown_symexp ce then ce 
    else (try lookup s va 
      with _ -> fresh_value ())


  (* McCarthy's Select Memory Axiom *) 
  | Lval(Mem(read_addr),NoOffset) -> 
    let read_addr = eval s read_addr in 
    let rec select lst = match lst with
    | [] -> fresh_value () 
    | (written_addr, written_value) :: earlier_writes -> begin 
      match decide s (BinOp(Eq,read_addr, written_addr, TInt(IInt,[]))) with
      | Must_Be_True -> written_value
      | Must_Be_False -> select earlier_writes 
      | Unknown -> fresh_value () 
    end 
    in 
    select s.mu 

  | UnOp(unop,ce,tau) -> UnOp(unop, eval s ce, tau) 
  | BinOp(bop,ce1,ce2,tau) -> begin
    match bop, (eval s ce1), (eval s ce2) with
    (* in a few cases we can compute this "in-line" without pushing it 
     * off to the theorem prover -- this is typically a worthwhile
     * optimization for scalability but might nott be worth it to you
     * in this homework *) 
    | PlusA, Const(CInt64(i1,ik1,_)), Const(CInt64(i2,_,_)) ->
      Const(CInt64(Int64.add i1 i2, ik1, None))
    | MinusA, Const(CInt64(i1,ik1,_)), Const(CInt64(i2,_,_)) ->
      Const(CInt64(Int64.sub i1 i2, ik1, None))
    | Mult, Const(CInt64(i1,ik1,_)), Const(CInt64(i2,_,_)) ->
      Const(CInt64(Int64.mul i1 i2, ik1, None))
    | Shiftlt, Const(CInt64(i1,ik1,_)), Const(CInt64(i2,_,_)) ->
      Const(CInt64(Int64.shift_left i1 (Int64.to_int i2), ik1, None))
    | Shiftrt, Const(CInt64(i1,ik1,_)), Const(CInt64(i2,_,_)) ->
      Const(CInt64(Int64.shift_right i1 (Int64.to_int i2), ik1, None))
    | Lt, Const(CInt64(i1,ik1,_)), Const(CInt64(i2,_,_)) ->
      se_of_bool (i1 < i2) 
    | Le, Const(CInt64(i1,ik1,_)), Const(CInt64(i2,_,_)) ->
      se_of_bool (i1 <= i2) 
    | Gt, Const(CInt64(i1,ik1,_)), Const(CInt64(i2,_,_)) ->
      se_of_bool (i1 > i2) 
    | Ge, Const(CInt64(i1,ik1,_)), Const(CInt64(i2,_,_)) ->
      se_of_bool (i1 >= i2) 
    | x, y, z -> BinOp(x,y,z,tau)
    end 
  | CastE(_,ce) -> eval s ce
  | x -> x

(*
 * This procedure evaluates an instruction in a given state. It returns
 * either Some(new_state) if symex should continue or None if we should
 * stop. We stop after calling exit(1) or on division by 0 or somesuch.
 *)
let rec handle_instr (i:instr) (s:state) : (state option) = match i with
  | Set((Var(va),NoOffset),new_val,location) -> 
    let new_val = eval s new_val in
    let s = add_to_path s location in 
    (Some (assign s va new_val))
  | Set((Mem(ptr_exp),NoOffset),new_val,location) -> 
    let ptr_exp = eval s ptr_exp in 
    let new_val = eval s new_val in 
    let s = add_to_path s location in 
    let s = update_memory s (ptr_exp) (new_val) in 
    (* magic trick: if you are continuing on a path after *p = 5, then 
     * p must be non-null! *) 
    let s = assume s (BinOp(Ne,ptr_exp,(se_of_bool false),TInt(IInt,[]))) in
    (Some s)

  | Set(_,new_val,location) -> 
    debug "Warning: tricky assignment!\n" ;
    (* you might want to fix this *) 
    let s = add_to_path s location in 
    (Some (throw_away_state s ))

  | Call(retval_option, (Lval(Var(va),NoOffset)), args, loc) -> 
    if va.vname = "exit" then None 
    else begin 
    let s = add_to_path s loc in 
    (* you could be more precise here and have a special whitelist /
     * blacklist of functions that you care about. You could also restrict
     * attention to arguments that have pointer types. *)
    let is_syscall = not (Filename.is_relative va.vdecl.file) || 
      va.vdecl.file = "" (* or it is undeclared *) in 
    if is_syscall then begin
      (* check to see if the arguments are zero *) 
      List.iter (fun arg_exp ->
        match decide s (eval s arg_exp) with
        | Must_Be_False -> 
          debug_state s ; 
          let estr = Pretty.sprint 80 (defaultCilPrinter#pExp () arg_exp) in
          report_error s.path 
            "*POSSIBLE BUG* in call to %s (%s believed to be 0)\n" 
            va.vname estr 
        | Must_Be_True -> () 
        | Unknown -> 
          if not !be_certain then begin 
            debug_state s ; 
            let estr = Pretty.sprint 80 (defaultCilPrinter#pExp () arg_exp) in
            report_error s.path 
              "*POSSIBLE BUG* in call to %s (%s may be 0)\n" 
              va.vname estr 
          end 
      ) args ; 
    end ;
    (match retval_option with
    | None -> (Some s)
    | Some(lv) -> 
      (* we are intraprocedural, so we assume the function call can 
       * return anything *) 
      let fv = fresh_value ~va () in 
      (handle_instr (Set(lv,fv,loc)) s)
    ) 
    end 

  | Call(retval_option, function_ptr_exp, args, location) -> 
    debug "Warning: call through function pointer!\n" ;
    (* you might want to fix this *) 
    let s = add_to_path s location in 
    (Some (throw_away_state s))

  | Asm(_) -> (* we don't handle inline asm! *) (Some s)


(*************************************************************************** 
 * SECTION D -- Symbolic Path / Dataflow Analysis
 *
 * Here we look at the control flow of the program and symbolically execute
 * a number of paths through it. The special 'num_paths' variable limits
 * the number of double-branches and cyclic paths we will explore.
 *
 * As written this is an INTRA-procedural analysis, so we start afresh 
 * with an empty state at the beginning of each method. 
 ***************************************************************************)
(* 
 * The maximum number of paths that will be considered per function. 
 *)
let num_paths = ref 2 

let analyze_function fundec = 
  let body : block = fundec.sbody in 
  let stmts = body.bstmts in 
  match stmts with
  | [] -> (* nothing to do -- empty function *) ()
  | first_statement :: others -> begin 

    let initial_state = empty_state in 
    let state = ref initial_state in 
    List.iter (fun local ->
      (* locals start out as zero! *) 
      state := assign !state local (se_of_bool false) 
    ) fundec.slocals ; 
    List.iter (fun formal -> 
      (* formals start out undefined -- we are not context-sensitive *)
      state := assign !state formal (fresh_value ~va:formal ())
    ) fundec.sformals ;
    (* if we were smarter, we would make this two-pass and gather up all 
       of the globals and do something about them -- for example, const
       globals could be set to their actual values *)

    let paths_left = ref (!num_paths) in 
    let worklist = Queue.create () in
    let initial_state = !state in 
    Queue.add (initial_state, first_statement) worklist ;
    while not (Queue.is_empty worklist) do
      let state, stmt = Queue.take worklist in 

      (* debugging *) 
      if not !be_quiet then begin 
        let _ = Pretty.printf "At Stmt #%d: @[@!%a@]@!@!" 
          stmt.sid 
          defaultCilPrinter#pStmt stmt in 
        flush stdout ; 
        () 
      end ; 

      if already_visited state stmt then begin
        (* if we have already been here, this is a cyclic path! 
         * for now we'll just drop (= not explore) the rest of this path --
         * for more precision you may want to, for example, "spend" one of
         * the paths remaining to keep going around the loop another time
         * *) 
        () 
      end else begin
        let state = mark_as_visited state stmt in 
        match stmt.skind with 
        (* if it is bunch of "x = 5; y = 3" sort of things, just execute 
         * all of them and go on ... *) 
        | Instr(il) -> begin 
          let new_state_opt = List.fold_left (fun state_opt instr ->
            match state_opt with
            | None -> None
            | Some(state) -> handle_instr instr state
          ) (Some state) il in
          match new_state_opt, stmt.succs with
          | None, _ -> ()
          | _, [] -> 
            debug "Warning: No Successors!\n" ; () (* nowhere to go ... *) 
          | Some(new_state), somewhere :: _ -> 
            Queue.add (new_state, somewhere) worklist
        end 
            
          (* this is an intraprocedural analysis, so we stop at
           * returns -- they behave like exit() *) 
        | Return(exp_option, _) -> () 
            
          (* goto doesn't change the state, so it's not hard *) 
        | Goto(target_stmt_ref, loc) -> 
          let state = add_to_path state loc in 
          Queue.add (state, !target_stmt_ref) worklist

          (* basically just a { list of statements } *)  
        | Block(newb) -> begin 
            match newb.bstmts @ stmt.succs with
            | [] -> 
              debug "Warning: No Block Successors!\n"  
                (* mysterious empty block? *) 
            | first_in_new_block :: rest -> 
              Queue.add (state, first_in_new_block) worklist
        end 

        (* We will try to statically determine if the predicate must be
         * true along the path we have traveled. If it must be true or
         * false given our context we don't have to explore both paths.
         * This both saves resources and prevents false positives. *)
        | If(exp,then_block,else_block,loc) -> begin 
          let state = add_to_path state loc in 
          let e = eval state exp in
          (* In a case like "if (b) { } else { foo(); } ; bar(); ", if b is 
           * true we have to make sure that we pick the successor that
           * leads to bar(), not the successor that leads to foo(). *)
          let rec pick_first_a_not_in_b a b = match a with
          | [] -> []
          | first_a :: rest_a -> 
            if List.mem first_a b then 
              pick_first_a_not_in_b rest_a b 
            else
              [first_a]
          in 
          match decide state e with
          | Must_Be_True -> 
            (match then_block.bstmts @ 
                    (pick_first_a_not_in_b stmt.succs else_block.bstmts) with
             | [] -> () | first :: rest -> Queue.add (state, first) worklist) 
          | Must_Be_False -> 
            (match else_block.bstmts @ 
                    (pick_first_a_not_in_b stmt.succs then_block.bstmts) with
             | [] -> () | first :: rest -> Queue.add (state, first) worklist) 
          | Unknown -> begin
            let s_then = assume state e in 
            (match then_block.bstmts @ stmt.succs with
             | [] -> () | first :: rest -> Queue.add (s_then, first) worklist);
            if !paths_left > 0 then begin
              decr paths_left ; 
              let s_else = assume state (UnOp(LNot,e,TInt(IInt,[]))) in 
              (match else_block.bstmts @ stmt.succs with
               | [] -> () | first :: rest -> Queue.add (s_else,first) worklist) 
            end ; 
          end 
        end 

        (* The WHILE loop is, as always, the most complicated part. I
         * haven't really filled anything in here -- you may/must do it. *) 
        | Loop(loop_body,loc,_,_) -> begin
          let state = add_to_path state loc in 
          match loop_body.bstmts with
          | [] -> (* while (1) skip == this path ends *) ()
          | first :: rest -> Queue.add (state, first) worklist
        end 

        (* You don't have to deal with this junk. Huzzah. *) 
        | Switch(_) -> failwith "switch should have been removed earlier" 
        | Break(_) -> failwith "break should have been removed earlier" 
        | Continue(_) -> failwith "continue should have been removed earlier" 
        | TryFinally(_)
        | TryExcept(_) -> failwith "MSVC try/finally not supported" 
      end 

    done 
  end 

(* you don't have to do it function by function -- you could do an
 * interprocedural analysis *)
let analyze glob = match glob with
  | GFun(fundec, location) -> 
    let ok_to_analyze = 
      let source_file = location.file in 
      Filename.is_relative source_file 
      (* don't analyze stuff in /usr/include/stdio.h *) 
    in 
    if ok_to_analyze then begin 
      if !only_functions = [] || List.mem fundec.svar.vname !only_functions
        then analyze_function fundec 
    end 

  | _ -> () 
  (* you could look at global variables and whatnot if you wanted to *) 

(*************************************************************************** 
 * SECTION E -- main() and driver
 *
 * Nothing exciting happens here. Check out the list of command line
 * arguments -- you may find the alias analysis one handy if you're just
 * playing around. 
 ***************************************************************************)

(* 
 * For each local pointer variable in the given function, indicate if it
 * may alias each other local pointer variable. This does not account for
 * global variables or or whatnot fields but you could if you wanted to.
 *) 
let print_aliases glob = match glob with
  | GFun(fundec, location) -> 
    let formals = fundec.sformals in
    let locals = fundec.slocals in 
    let pointers = List.filter (fun v -> isPointerType v.vtype ) 
      (formals @ locals) in 
    let vars = Array.of_list pointers in 
    let num_vars = Array.length vars in 
    for i = 0 to pred num_vars do
      for j = i + 1 to pred num_vars do
        let v1 = vars.(i) in 
        let v2 = vars.(j) in 
        let e1 = Lval(Var(v1),NoOffset) in 
        let e2 = Lval(Var(v2),NoOffset) in 
        let ma = try Ptranal.may_alias e1 e2 with _ -> true in
        if ma then begin
          debug "May Alias: %s %s\n" v1.vname v2.vname 
        end 
      done 
    done 

  | _ -> () 


let main () =
  let print_cil_fun = ref [] in 
  let print_c_fun = ref [] in 
  let print_alias = ref [] in 
  let files_to_process = ref [] in 
  Cil.printCilAsIs := true ; 

  let argDesc = [ 
    "--quiet", Arg.Set be_quiet, "\tonly emit output about errors" ;
    "--debug-tp", Arg.Set debug_tp, "\tdebug theorem prover interaction" ; 
    "--only-function", Arg.String (fun x ->
      only_functions := x :: !only_functions), "\tonly analyze the given functions"; 
    "--print-alias", Arg.String( fun x -> 
      print_alias := x :: !print_alias),"\tprint out the local aliases for the given function" ; 
    "--print-cil" , Arg.String ( fun x ->
      print_cil_fun := x :: !print_cil_fun),"\tprint out the CIL intermediate code for the given function" ; 
    "--print-c" , Arg.String ( fun x ->
      print_c_fun := x :: !print_c_fun),"\tprint out the C code for the given function" ; 
    "--max-errors", Arg.Int (fun x -> stop_at_error := x),
      "\texit after reporting this many errors"; 
    "--be-certain", Arg.Set be_certain, "\tonly report errors that are very likely to be real" ;
    "--paths", Arg.Int (fun x -> num_paths := x), 
      "\tmaximum number of paths to consider per function" ; 
  ] in 

  let handle_file str = files_to_process := str :: !files_to_process in 

  Arg.parse argDesc handle_file "Usage:\n" ; 

  List.iter (fun file ->
    debug "%s : processing\n" file ;
    let cil_ast = Frontc.parse file () in
    debug "%s : parsed \n" file ; 

    ignore (Simplemem.simplemem cil_ast) ; 

    debug "%s : simplified memory expressions\n" file ; 

    iterGlobals cil_ast Simplify.doGlobal ; 

    debug "%s : simplified to three-address code\n" file ; 

    ignore (Partial.calls_end_basic_blocks cil_ast) ; 
    ignore (Partial.globally_unique_vids cil_ast) ; 
    iterGlobals cil_ast (fun glob -> match glob with 
    | GFun(fd,_) -> prepareCFG fd ; ignore (computeCFGInfo fd true)
    | _ -> () 
    ) ; 
    debug "%s : made more like a CFG\n" file ;  

    ignore (Ptranal.analyze_file cil_ast); 
    ignore (Ptranal.compute_results false); 
    debug "%s : analyzed aliases\n" file ; 

    iterGlobals cil_ast (fun glob -> match glob with
    | GFun(fundec, location) -> 
      let function_name = fundec.svar.vname in 
      if List.mem function_name !print_cil_fun then begin
        ignore (Pretty.printf "%a\n" plainCilPrinter#pGlobal glob) ; 
      end ;
      if List.mem function_name !print_c_fun then begin
        ignore (Pretty.printf "%a\n" defaultCilPrinter#pGlobal glob) ; 
      end ;
      if List.mem function_name !print_alias then begin
        print_aliases glob
      end ;
      analyze glob 

    | _ -> () 
    ) ;

  ) !files_to_process ;

  printf "\n%d errors reported.\n" !error_count ; 

;;
main () ;; 

(* It took WRW ~10 hours to write this (Feb 15, 10:20 to Feb 15, 20:20). 
 * Which goes a ways toward explaining why it is buggy. *)
